/**
 * @file geode-fb.h
 * @brief Native driver for the Geode GX1 processor and chipset
 * @author Juergen Beisert <jbeisert@netscape.net>
 *
 * This driver bases on Xorg's fbdev
 */

/**
 * @brief FIXME
 **/
#define GEODEDEV_VERSION	4000
/**
 * @brief FIXME
 **/
#define GEODEDEV_NAME		"FBDEV"
/**
 * @brief FIXME
 **/
#define GEODEDEV_DRIVER_NAME	"geodegx1_drv"
/**
 * @brief Major release number of this driver
 * The Xorg server outputs this information when it loads this module
 **/
#define GEODEDEV_MAJOR_VERSION	1
/**
 * @brief Minor release number of this driver
 * The Xorg server outputs this information when it loads this module
 **/
#define GEODEDEV_MINOR_VERSION	3
/**
 * @brief Patch level of this driver
 * The Xorg server outputs this information when it loads this module
 **/
#define GEODEDEV_PATCH_LEVEL	0

extern Bool FBDevPreInit(ScrnInfoRec*, int);
extern Bool FBDevProbe(DriverPtr, int);
extern const OptionInfoRec * FBDevAvailableOptions(int, int);
extern Bool FBDevDriverFunc(ScrnInfoPtr, xorgDriverFuncOp, pointer);
extern int geode_gx1_register_symbols(void);

#ifdef DEBUG
# define TRACE_ENTER(str)       ErrorF("%s: %s entered\n",GEODEDEV_DRIVER_NAME,str)
# define TRACE_EXIT(str)        ErrorF("%s: %s done\n",GEODEDEV_DRIVER_NAME,str)
# define TRACE(str)             xf86ErrorF("%s trace: %s\n",GEODEDEV_DRIVER_NAME,str)
#else
# define TRACE_ENTER(str)
# define TRACE_EXIT(str)
# define TRACE(str)
#endif

/*
 * we need this for some structures in our private data structure
 */
#include <linux/fb.h>

/*
 * We are relying on a special framebuffer driver
 */
#include <geode_gx1_fb.h>

/**
 * @brief private data handling
 **/
typedef struct
{
	/* some flags controlled by settings in the xorg.conf */
	int				rotate;		/*<< != 0 is some rotations should be used */
	int				shadowFB;	/*<< TRUE if shadow FB should be enabled */
	int				accel;		/*<< TRUE if acceleration should be used */
	int				compression;	/*<< TRUE if compression should be enabled */

	CloseScreenProcPtr	CloseScreen;
	void	( *PointerMoved ) ( int index, int x, int y );
	EntityInfoPtr	pEnt;
	/* DGA info */
	/* DGAModePtr	pDGAMode; */
	/* int	nDGAMode; */
	OptionInfoRec	*Options;

	char*				device;
	int				fd;
	unsigned int			fboff;
	/* char*				mmio; */
	/* unsigned int			mmio_len; */

	/* current hardware state */
	struct fb_fix_screeninfo	fix;
	struct fb_var_screeninfo	var;

	/* saved video mode */
	/* struct fb_var_screeninfo	saved_var; */

	/* FIXME: unused??? [geert] */
	/* struct fb_cmap			saved_cmap; */
	/* unsigned short			*saved_red;
	unsigned short			*saved_green;
	unsigned short			*saved_blue; */

	unsigned long			pagesize;	/*<< pagesize of this system */
/*
 * for future use
 */
	unsigned long dc_registers;	/*<< physical address of GX1's DC registers */
	unsigned long gp_registers;	/*<< physical address of GX1's graphics pipeline registers */
	unsigned long scratch_pad_ram;	/*<< physical address of GX1's scratch pad SRAM */
	int scratch_offset;		/*<< offset of the first useable byte in scratch pad SRAM */
	unsigned long video_mem;	/*<< physical base address of the video memory */

	/* compression support */
	unsigned long	cbuffer_offset;	/*<< offset for compression buffer start */
	unsigned long	cbuffer_ldelta;	/*<< line delta for compression buffer */
	unsigned long	cbuffer_lsize;	/*<< line size in bytes for the compression buffer */

	/* for easier cleanup */
	void *cursorPtr;		/*<< hardware cursor info */
	void *AccelInfoRec;		/*<< data to handle acceleration */

/* the following pointers are for the CPU */
	void *pci_io;			/*<< virtual base address of the PCI part */
	void *vmem_io;			/*<< virtual base address of the video memory */
	void *dc_io;			/*<< virtual base address of GX1's DC registers */
	void *gp_io;			/*<< virtual base address of GX1's GP registers */
	void *scratch_io;		/*<< virtual base address of GX1's Scratch Pad RAM */

	unsigned char *fbstart;		/*<< virtual top left of screen in framebuffer */
	unsigned int fbmem_len;		/*<< size of videomemory in bytes */
	int real_line_length;		/*<< real line length in bytes (so called line delta) */

/* hardware cursor handling */
	int use_second_cursor;		/*<< 0=cursor_1 , 1=cursor_2 should be displayed */
	unsigned long cursor_yoffset;	/*<< current cursor Y offset (for hiding at top) */

	unsigned long *cursor_1;	/*<< 256 bytes of offscreen buffer for cursor image */
	unsigned long *cursor_2;	/*<< 256 bytes of offscreen buffer for cursor image */

/* these points into video memory (offscreen) */
	unsigned long cursor_offset_1;	/*<< offset of cursor image 1 in video memory */
	unsigned long cursor_offset_2;	/*<< offset of cursor image 2 in video memory */

/* data to handle acceleration internally */
	int geode_not_native;		/*<< true if other colour depth than 16 */

	int GFXsourceFlags;
	int GFXpatternFlags;
	unsigned int GFXsavedColour;	/*<< colour for solid fills and lines */
	unsigned int GFXsavedRop;
	unsigned int GFXusesDstData;
	unsigned int GeodebufferWidthPixels;
	int Geodebb0Base,Geodebb1Base;	/*<< Offsets into Scratch Pad RAM */
	int GeodeTransparent;
	unsigned long GeodeTransColour;	/*<< Transparent colour while BLT */
	unsigned int Geode_blt_mode;
	unsigned int Geode_vector_mode;
	unsigned int Geode_buffer_width;/*<< BLT buffer width for BLT operations */
	int yshift;
}
GeodeDevRec;

/**
 * @brief extract out private data pointer
 **/
#define GEODEDEVPTR(p) ((GeodeDevRec *)((p)->driverPrivate))

/* FIXME use winSetScreenPriv() and winGetScreenPriv() instead */

/* from geode_cursor.c */
extern int geode_hw_cursor_init(ScreenRec*);
extern int geode_hw_cursor_exit(ScreenRec*);

/* from geode_accel.c */
extern int geode_init_acceleration(ScreenRec*);
extern int geode_exit_acceleration(ScreenRec*);
extern int geode_setup_blitbuffer_pointers(GeodeDevRec*,int,int);
/* extern int geode_setup_gamma_table(GeodeDevRec*,LOCO*); currently not used */
extern int geode_enable_compression(GeodeDevRec*);

/**
 * @brief Window size to map
 **/
#define GEODE_DC_WINDOW_SIZE 0x100
/**
 * @brief Graphics Pipeline window size to map
 **/
#define GEODE_GP_WINDOW_SIZE 0x200

/**
 * @brief Scratch pad area is 4kiB max
 **/
#define GEODE_SCRATCH_WINDOW_SIZE 0x1000

/* Display controller registers */

#define DC_PAL_ADDRESS		0x70
#define DC_PAL_DATA		0x74

#define DC_UNLOCK		0x00
#  define DC_UNLOCK_CODE		0x00004758

#define DC_GENERAL_CFG		0x04
#  define DC_GCFG_DFLE			0x00000001
#  define DC_GCFG_CURE			0x00000002
#  define DC_GCFG_VCLK_DIV		0x00000004
#  define DC_GCFG_PLNO			0x00000004
#  define DC_GCFG_PPC			0x00000008
#  define DC_GCFG_CMPE			0x00000010
#  define DC_GCFG_DECE			0x00000020
#  define DC_GCFG_DCLK_MASK		0x000000C0
#  define DC_GCFG_DCLK_DIV_1		0x00000080
#  define DC_GCFG_DFHPSL_MASK		0x00000F00
#  define DC_GCFG_DFHPSL_POS			 8
#  define DC_GCFG_DFHPEL_MASK		0x0000F000
#  define DC_GCFG_DFHPEL_POS			12
#  define DC_GCFG_CIM_MASK		0x00030000
#  define DC_GCFG_CIM_POS			16
#  define DC_GCFG_FDTY			0x00040000
#  define DC_GCFG_RTPM			0x00080000
#  define DC_GCFG_DAC_RS_MASK		0x00700000
#  define DC_GCFG_DAC_RS_POS			20
#  define DC_GCFG_CKWR			0x00800000
#  define DC_GCFG_LDBL			0x01000000
#  define DC_GCFG_DIAG			0x02000000
#  define DC_GCFG_CH4S			0x04000000
#  define DC_GCFG_SSLC			0x08000000
#  define DC_GCFG_VIDE			0x10000000
#  define DC_GCFG_VRDY			0x20000000
#  define DC_GCFG_DPCK			0x40000000
#  define DC_GCFG_DDCK			0x80000000

#define DC_TIMING_CFG		0x08
#  define DC_TCFG_FPPE			0x00000001
#  define DC_TCFG_HSYE			0x00000002
#  define DC_TCFG_VSYE			0x00000004
#  define DC_TCFG_BLKE			0x00000008
#  define DC_TCFG_DDCK			0x00000010
#  define DC_TCFG_TGEN			0x00000020
#  define DC_TCFG_VIEN			0x00000040
#  define DC_TCFG_BLNK			0x00000080
#  define DC_TCFG_CHSP			0x00000100
#  define DC_TCFG_CVSP			0x00000200
#  define DC_TCFG_FHSP			0x00000400
#  define DC_TCFG_FVSP			0x00000800
#  define DC_TCFG_FCEN			0x00001000
#  define DC_TCFG_CDCE			0x00002000
#  define DC_TCFG_PLNR			0x00002000
#  define DC_TCFG_INTL			0x00004000
#  define DC_TCFG_PXDB			0x00008000
#  define DC_TCFG_BKRT			0x00010000
#  define DC_TCFG_PSD_MASK		0x000E0000
#  define DC_TCFG_PSD_POS			17
#  define DC_TCFG_DDCI			0x08000000
#  define DC_TCFG_SENS			0x10000000
#  define DC_TCFG_DNA			0x20000000
#  define DC_TCFG_VNA			0x40000000
#  define DC_TCFG_VINT			0x80000000

#define DC_OUTPUT_CFG		0x0C
#  define DC_OCFG_8BPP			0x00000001
#  define DC_OCFG_555			0x00000002
#  define DC_OCFG_PCKE			0x00000004
#  define DC_OCFG_FRME			0x00000008
#  define DC_OCFG_DITE			0x00000010
#  define DC_OCFG_2PXE			0x00000020
#  define DC_OCFG_2XCK			0x00000040
#  define DC_OCFG_2IND			0x00000080
#  define DC_OCFG_34ADD			0x00000100
#  define DC_OCFG_FRMS			0x00000200
#  define DC_OCFG_CKSL			0x00000400
#  define DC_OCFG_PRMP			0x00000800
#  define DC_OCFG_PDEL			0x00001000
#  define DC_OCFG_PDEH			0x00002000
#  define DC_OCFG_CFRW			0x00004000
#  define DC_OCFG_DIAG			0x00008000

#define DC_CB_ST_OFFSET		0x14
#define DC_CURS_ST_OFFSET	0x18
#define DC_VID_ST_OFFSET	0x1c

#define DC_LINE_DELTA		0x24
# define DC_LINE_DELTA_FB_MASK 0x0007ff
# define DC_LINE_DELTA_CB_MASK 0x7ff000

#define DC_BUF_SIZE		0x28
# define DC_BUF_SIZE_FB_MASK	0x000001ff
# define DC_BUF_SIZE_CB_MASK	0x0000fe00
# define DC_BUF_SIZE_VD_MASK	0x3fff0000

#define DC_CURSOR_X		0x50
#define DC_CURSOR_Y		0x58

#define DC_PAL_ADDRESS		0x70
#define DC_PAL_DATA		0x74

#define GP_DST_XCOOR		0x0000
#define GP_DST_YCOOR		0x0002
#define GP_WIDTH		0x0004
#define GP_HEIGHT		0x0006
#define GP_SRC_XCOOR		0x0008
#define GP_SRC_YCOOR		0x000A
#define GP_SRC_COLOR_0		0x000C
#define GP_SRC_COLOR_1		0x000E
#define GP_PAT_COLOR_0		0x0010
#define GP_PAT_COLOR_1		0x0012
#define GP_PAT_COLOR_2		0x0014
#define GP_PAT_COLOR_3		0x0016
#define GP_PAT_DATA_0		0x0020
#define GP_PAT_DATA_1		0x0024
#define GP_PAT_DATA_2		0x0028
#define GP_PAT_DATA_3		0x002C

#define GP_VECTOR_LENGTH	0x0004	/* vector length            */
#define GP_INIT_ERROR		0x0006	/* vector initial error     */
#define GP_AXIAL_ERROR		0x0008	/* axial error increment    */
#define GP_DIAG_ERROR		0x000A	/* diagonal error increment */

#define GP_RASTER_MODE		0x0100
# define RM_PAT_DISABLE		0x0000	/* pattern is disabled      */
# define RM_PAT_MONO		0x0100	/* 1BPP pattern expansion   */
# define RM_PAT_DITHER		0x0200	/* 2BPP pattern expansion   */
# define RM_PAT_COLOR		0x0300	/* 8BPP or 16BPP pattern    */
# define RM_PAT_MASK		0x0300	/* mask for pattern mode    */

# define RM_PAT_TRANSPARENT	0x0400	/* transparent 1BPP pattern */
# define RM_SRC_TRANSPARENT	0x0800	/* transparent 1BPP source  */

# define GP_VECTOR_MODE		0x0104

/* --- */

# define GP_BLIT_MODE		0x0108

/* "GP_BLIT_MODE" BIT DEFINITIONS */

# define BM_READ_SRC_NONE	0x0000	/* source foreground color      */
# define BM_READ_SRC_FB		0x0001	/* read source from FB          */
# define BM_READ_SRC_BB0	0x0002	/* read source from BB0         */
# define BM_READ_SRC_BB1	0x0003	/* read source from BB1         */
# define BM_READ_SRC_MASK	0x0003	/* read source mask             */

# define BM_READ_DST_NONE	0x0000	/* no destination data          */
# define BM_READ_DST_BB0	0x0008	/* destination from BB0         */
# define BM_READ_DST_BB1	0x000C	/* destination from BB1         */
# define BM_READ_DST_FB0	0x0010	/* dest from FB (store BB0)     */
# define BM_READ_DST_FB1	0x0014	/* dest from FB (store BB1)     */
# define BM_READ_DST_MASK	0x001C	/* read destination mask        */

# define BM_WRITE_FB		0x0000	/* write to framebuffer         */
# define BM_WRITE_MEM		0x0020	/* write to memory              */
# define BM_WRITE_MASK		0x0020	/* write mask                   */

# define BM_SOURCE_COLOR	0x0000	/* source is 8BPP or 16BPP      */
# define BM_SOURCE_EXPAND	0x0040	/* source is 1BPP               */
# define BM_SOURCE_TEXT		0x00C0	/* source is 1BPP text          */
# define BM_SOURCE_MASK		0x00C0	/* source mask                  */

# define BM_REVERSE_Y		0x0100	/* reverse Y direction          */

/* --- */

#define GP_BLIT_STATUS		0x010C

# define BS_SCREENW_MASK	0x0600
# define BS_COLOR_M		0x0100
# define BS_FLUSH		0x0080	/* ??? */
# define BS_BLIT_PENDING	0x0004
# define BS_PIPELINE_BUSY	0x0002
# define BS_BLIT_BUSY		0x0001

/* "GP_VECTOR_MODE" BIT DEFINITIONS */

#define VM_X_MAJOR		0x0000	/* X major vector                       */
#define VM_Y_MAJOR		0x0001	/* Y major vector                       */
#define VM_MAJOR_INC		0x0002	/* positive major axis step */
#define VM_MINOR_INC		0x0004	/* positive minor axis step */
#define VM_READ_DST_FB		0x0008	/* read destination data        */

/* CS5530 Video device registers (PCI space)*/

#define CS5530_VIDEO_CONFIG	0x0000
#  define CS5530_VCFG_VID_EN			0x00000001
#  define CS5530_VCFG_VID_REG_UPDATE		0x00000002
#  define CS5530_VCFG_VID_INP_FORMAT		0x0000000C
#  define CS5530_VCFG_8_BIT_4_2_0		0x00000004
#  define CS5530_VCFG_16_BIT_4_2_0		0x00000008
#  define CS5530_VCFG_GV_SEL			0x00000010
#  define CS5530_VCFG_CSC_BYPASS		0x00000020
#  define CS5530_VCFG_X_FILTER_EN		0x00000040
#  define CS5530_VCFG_Y_FILTER_EN		0x00000080
#  define CS5530_VCFG_LINE_SIZE_LOWER_MASK	0x0000FF00
#  define CS5530_VCFG_INIT_READ_MASK		0x01FF0000
#  define CS5530_VCFG_EARLY_VID_RDY		0x02000000
#  define CS5530_VCFG_LINE_SIZE_UPPER		0x08000000
#  define CS5530_VCFG_4_2_0_MODE		0x10000000
#  define CS5530_VCFG_16_BIT_EN			0x20000000
#  define CS5530_VCFG_HIGH_SPD_INT		0x40000000

#define CS5530_DISPLAY_CONFIG	0x0004
#  define CS5530_DCFG_DIS_EN			0x00000001
#  define CS5530_DCFG_HSYNC_EN			0x00000002
#  define CS5530_DCFG_VSYNC_EN			0x00000004
#  define CS5530_DCFG_DAC_BL_EN			0x00000008
#  define CS5530_DCFG_DAC_PWR_EN		0x00000020
#  define CS5530_DCFG_FP_PWR_EN			0x00000040
#  define CS5530_DCFG_FP_DATA_EN		0x00000080
#  define CS5530_DCFG_CRT_HSYNC_POL		0x00000100
#  define CS5530_DCFG_CRT_VSYNC_POL		0x00000200
#  define CS5530_DCFG_FP_HSYNC_POL		0x00000400
#  define CS5530_DCFG_FP_VSYNC_POL		0x00000800
#  define CS5530_DCFG_XGA_FP			0x00001000
#  define CS5530_DCFG_FP_DITH_EN		0x00002000
#  define CS5530_DCFG_CRT_SYNC_SKW_MASK		0x0001C000
#  define CS5530_DCFG_CRT_SYNC_SKW_INIT		0x00010000
#  define CS5530_DCFG_PWR_SEQ_DLY_MASK		0x000E0000
#  define CS5530_DCFG_PWR_SEQ_DLY_INIT		0x00080000
#  define CS5530_DCFG_VG_CK			0x00100000
#  define CS5530_DCFG_GV_PAL_BYP		0x00200000
#  define CS5530_DCFG_DDC_SCL			0x00400000
#  define CS5530_DCFG_DDC_SDA			0x00800000
#  define CS5530_DCFG_DDC_OE			0x01000000
#  define CS5530_DCFG_16_BIT_EN			0x02000000

#define CS5530_PALETTE_ADDRESS	0x001C
#define CS5530_PALETTE_DATA	0x0020
/**
 * @brief FIXME
 **/
#define ABS(_val1, _val2) (((_val1) > (_val2)) ? ((_val1)-(_val2)) : ((_val2) - (_val1)))

/**
 * @brief read a 32 bit word volatile from IO space
 * @param[in] base Virtual base
 * @param[in] reg Real offset
 *
 * @return Value read
 **/
static __inline__ unsigned long READ_REG32(void *base,int reg)
{
	volatile unsigned long *adr = (volatile unsigned long*)(base + reg);
	return *adr;
}

/**
 * @brief read a 16 bit word volatile from IO space
 * @param[in] base Virtual base
 * @param[in] reg Real offset
 *
 * @return Value read
 **/
static __inline__ unsigned short READ_REG16(void *base,int reg)
{
	volatile unsigned short *adr = (volatile unsigned short*)(base + reg);
	return *adr;
}

/**
 * @brief write a 32 bit word volatile into IO space
 * @param[in] base Virtual base
 * @param[in] reg Real offset
 * @param[in] value Value to write
 **/
static __inline__ void WRITE_REG32(void *base,int reg,unsigned long value)
{
	volatile unsigned long *adr = (volatile unsigned long*)(base + reg);
	*adr=value;
}

/**
 * @brief write a 16 bit word volatile into IO space
 * @param[in] base Virtual base
 * @param[in] reg Real offset
 * @param[in] value Value to write
 **/
static __inline__ void WRITE_REG16(void *base,int reg,unsigned short value)
{
	volatile unsigned short *adr = (volatile unsigned short*)(base + reg);
	*adr=value;
}

/* end of file geode-fb.h */
