/**
 * @file geode_dga.c
 * @brief "Direct Graphics Architecture" support for applications that wants
 * to share the framebuffer with the X-Server
 *
 * TODO: To be included only if the server supports DGA!
 *
 **/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* all driver need this */
#include "xf86.h"

#include "dgaproc.h"

#if 0
/***********************************************************************
 * DGA stuff
 ***********************************************************************/
static Bool FBDevDGAOpenFramebuffer( ScrnInfoPtr pScrn, char **DeviceName,
                                     unsigned char **ApertureBase,
                                     int *ApertureSize, int *ApertureOffset,
                                     int *flags );
static Bool FBDevDGASetMode( ScrnInfoPtr pScrn, DGAModePtr pDGAMode );
static void FBDevDGASetViewport( ScrnInfoPtr pScrn, int x, int y, int flags );

/**
 * @brief FIXME
 * @param[in] pScrn
 * @param[out] DeviceName
 * @param[out] ApertureBase
 * @param[out] ApertureSize
 * @param[out] ApertureOffset
 * @param[out] flags
 **/
static Bool FBDevDGAOpenFramebuffer( ScrnInfoPtr pScrn, char **DeviceName,
                         unsigned char **ApertureBase, int *ApertureSize,
                         int *ApertureOffset, int *flags )
{
	TRACE_ENTER(__FUNCTION__);
					*DeviceName = NULL;		/* No special device */
	*ApertureBase = ( unsigned char * ) ( pScrn->memPhysBase );
	*ApertureSize = pScrn->videoRam;
	*ApertureOffset = pScrn->fbOffset;
	*flags = 0;

	return TRUE;
}

/**
 * @brief FIXME
 * @param[in] pScrn
 * @param[in] pDGAMode
 **/
static Bool FBDevDGASetMode( ScrnInfoPtr pScrn, DGAModePtr pDGAMode )
{
	DisplayModePtr pMode;
	int scrnIdx = pScrn->pScreen->myNum;
	int frameX0, frameY0;

	TRACE_ENTER(__FUNCTION__);
	if ( pDGAMode ) {
		pMode = pDGAMode->mode;
		frameX0 = frameY0 = 0;
} else {
		if ( !( pMode = pScrn->currentMode ) )
			return TRUE;

		frameX0 = pScrn->frameX0;
		frameY0 = pScrn->frameY0;
}

	if ( !( *pScrn->SwitchMode ) ( scrnIdx, pMode, 0 ) )
		return FALSE;
	( *pScrn->AdjustFrame ) ( scrnIdx, frameX0, frameY0, 0 );

	TRACE_EXIT(__FUNCTION__);
	return TRUE;
}

static void FBDevDGASetViewport( ScrnInfoPtr pScrn, int x, int y, int flags )
{
	TRACE_ENTER(__FUNCTION__);
	( *pScrn->AdjustFrame ) ( pScrn->pScreen->myNum, x, y, flags );
}

static int FBDevDGAGetViewport( ScrnInfoPtr pScrn )
{
	TRACE_ENTER(__FUNCTION__);
	return ( 0 );
}

static DGAFunctionRec FBDevDGAFunctions =
{
	FBDevDGAOpenFramebuffer,
	NULL,        /* CloseFramebuffer */
	FBDevDGASetMode,
	FBDevDGASetViewport,
	FBDevDGAGetViewport,
	NULL,        /* Sync */
	NULL,        /* FillRect */
	NULL,        /* BlitRect */
	NULL,        /* BlitTransRect */
};

static void FBDevDGAAddModes( ScrnInfoPtr pScrn )
{
	FBDevRec * fPtr = FBDEVPTR( pScrn );
	DisplayModePtr pMode = pScrn->modes;
	DGAModePtr pDGAMode;

	TRACE_ENTER(__FUNCTION__);
	do {
		pDGAMode = xrealloc( fPtr->pDGAMode,
		                     ( fPtr->nDGAMode + 1 ) * sizeof( DGAModeRec ) );
		if ( !pDGAMode )
			break;

		fPtr->pDGAMode = pDGAMode;
		pDGAMode += fPtr->nDGAMode;
		( void ) memset( pDGAMode, 0, sizeof( DGAModeRec ) );

		++fPtr->nDGAMode;
		pDGAMode->mode = pMode;
		pDGAMode->flags = DGA_CONCURRENT_ACCESS | DGA_PIXMAP_AVAILABLE;
		pDGAMode->byteOrder = pScrn->imageByteOrder;
		pDGAMode->depth = pScrn->depth;
		pDGAMode->bitsPerPixel = pScrn->bitsPerPixel;
		pDGAMode->red_mask = pScrn->mask.red;
		pDGAMode->green_mask = pScrn->mask.green;
		pDGAMode->blue_mask = pScrn->mask.blue;
		pDGAMode->visualClass = pScrn->bitsPerPixel > 8 ?
		                        TrueColor : PseudoColor;
		pDGAMode->xViewportStep = 1;
		pDGAMode->yViewportStep = 1;
		pDGAMode->viewportWidth = pMode->HDisplay;
		pDGAMode->viewportHeight = pMode->VDisplay;

		/*
		 * FIXME Is the real mode currently active? If not,
		 * this value is wrong!
		*/
		pDGAMode->bytesPerScanline = fPtr->real_line_length;

		pDGAMode->imageWidth = pMode->HDisplay;
		pDGAMode->imageHeight = pMode->VDisplay;
		pDGAMode->pixmapWidth = pDGAMode->imageWidth;
		pDGAMode->pixmapHeight = pDGAMode->imageHeight;
		pDGAMode->maxViewportX = pScrn->virtualX -
		                         pDGAMode->viewportWidth;
		pDGAMode->maxViewportY = pScrn->virtualY -
		                         pDGAMode->viewportHeight;

		pDGAMode->address = fPtr->fbstart;

		pMode = pMode->next;
} while ( pMode != pScrn->modes );
	TRACE_EXIT(__FUNCTION__);
}

static Bool FBDevDGAInit( ScrnInfoPtr pScrn, ScreenPtr pScreen )
{
	FBDevRec * fPtr = FBDEVPTR( pScrn );

	TRACE_ENTER(__FUNCTION__);
	if ( pScrn->depth < 8 )
		return FALSE;

	if ( !fPtr->nDGAMode )
		FBDevDGAAddModes( pScrn );

	TRACE_EXIT(__FUNCTION__);
	return ( DGAInit( pScreen, &FBDevDGAFunctions,
			fPtr->pDGAMode, fPtr->nDGAMode ) );
}
#endif
